@@ -1,4 +1,4 @@
-#define VERSION "1.53"
+#define VERSION "1.54"
 
 /*
  *  units, a program for units conversion
@@ -47,6 +47,7 @@
 #endif /* !strchr */
 
 char *strtok(); 
+double strtod();
 
 #ifdef READLINE
 #  include <readline/readline.h>
@@ -75,6 +76,7 @@ int quiet = 0;                  /* Flag for supressing prompting (-q option) */
 int unitcheck = 0;              /* Flag for unit checking, set to 1 for 
                                    regular check, set to 2 for verbose check */
 int verbose = 0;                /* Flag for verbose operation */
+int strictconvert = 0;          /* Strict conversion (disables reciprocals) */
 char *userfile = NULL;          /* User specified units file name */
 char *progname="units";         /* Used in error messages */
 
@@ -126,7 +128,7 @@ struct unittype {
 /* Increases the buffer by BUFGROW bytes and leaves the new pointer in buf
    and the new buffer size in bufsize. */
 
-#define BUFGROW 10/*24*/  /* For debugging purposes, buffers grow very slowly*/
+#define BUFGROW 256
 
 void
 growbuffer(char **buf, int *bufsize)
@@ -550,34 +552,28 @@ zeroerror()
 
 
 /* 
-   Finds the length of the largest prefix of the argument which
-   is a number.  A number has the form of a sequence of combinations
-   that are regular numbers separated by '|' characters.  No spaces  
-   are allowed and the string must not start or end with '|'.
+   Finds the length of the largest prefix of the argument which is a
+   number.  A number is a sequence of C doubles separated by '|'
+   characters.  No spaces are allowed and the string must not start or
+   end with '|'.
 */
 
 
 int
 numlength(char *str)
 {
-   float dummy;
-   int length,answer,backup;
+   char *trystart, *tryend, *numend;
 
-   backup=0;
-   answer=0;
-   for(;;){
-     if (1 != sscanf(str,"%f%n",&dummy,&length)) {
-       answer-=backup;
-       break;
-     }
-     answer+=length;
-     str+=length;
-     if (*str != '|') break;
-     str++;
-     answer++;
-     backup = 1;
+   numend = str;    /* numend marks the end of number so far */
+   trystart = str;
+   for (;;) {
+     strtod(trystart, &tryend);
+     if (tryend == trystart) break;
+     numend = tryend;
+     if (*numend != '|') break; /* If there's no '|' number is over */
+     trystart = numend + 1;     /* Otherwise check for number after '|'*/
    }
-   return answer;
+   return numend - str;
 }
 
 
@@ -641,7 +637,12 @@ cancelunit(struct unittype *theunit)
 
    Sometimes the returned pointer will be a pointer to the special
    buffer created to hold the data.  This buffer grows as needed during 
-   program execution.  */
+   program execution.  
+
+   Don't pass the output of lookupunit() back into the function again,
+   however, because the buffer may be overwritten producing incorrect
+   results.
+*/
 
 static int bufsize=0;
 static char *buffer;  /* buffer for lookupunit answers with prefixes */
@@ -662,22 +663,6 @@ lookupunit(char *unit,int prefixok)
    if (uptr = ulookup(unit))
       return uptr->value;
 
-   if (prefixok && (pfxptr = plookup(unit))) {
-      copy=unit;
-      copy += pfxptr->len;
-      if (!strlen(copy) || lookupunit(copy,0)) {
-         char *tempbuf;
-         while (strlen(pfxptr->value)+strlen(copy)+2 > bufsize){
-            growbuffer(&buffer, &bufsize);
-         }
-         tempbuf = strdup(copy);   /* copy might point into buffer */
-	 strcpy(buffer, pfxptr->value);
-	 strcat(buffer, " ");
-	 strcat(buffer, tempbuf);
-         free(tempbuf);
-	 return buffer;
-      }
-   }
    if (strlen(unit)>2 && unit[strlen(unit) - 1] == 's') {
       copy = dupstr(unit);
       copy[strlen(copy) - 1] = 0;
@@ -702,6 +687,22 @@ lookupunit(char *unit,int prefixok)
       }
       free(copy);
    }
+   if (prefixok && (pfxptr = plookup(unit))) {
+      copy=unit;
+      copy += pfxptr->len;
+      if (!strlen(copy) || lookupunit(copy,0)) {
+         char *tempbuf;
+         while (strlen(pfxptr->value)+strlen(copy)+2 > bufsize){
+            growbuffer(&buffer, &bufsize);
+         }
+         tempbuf = dupstr(copy);   /* copy might point into buffer */
+	 strcpy(buffer, pfxptr->value);
+	 strcat(buffer, " ");
+	 strcat(buffer, tempbuf);
+         free(tempbuf);
+	 return buffer;
+      }
+   }
    return 0;
 }
 
@@ -741,7 +742,7 @@ char *
 removepadding(char *in)
 {
    in += strspn(in," \t\n");
-   while(strchr(" \t\n",in[strlen(in)-1])) in[strlen(in)-1]=0;
+   while(*in && strchr(" \t\n",in[strlen(in)-1])) in[strlen(in)-1]=0;
    return in;
 }
 
@@ -754,9 +755,8 @@ int
 addnumber(struct unittype *theunit, int flip, char  *item)
 {            
   double number;
-  int len;
   
-  sscanf(item,"%lf%n",&number,&len);
+  number = strtod(item, &item);
   if (!number) {
      zeroerror();
      return  1;
@@ -765,14 +765,13 @@ addnumber(struct unittype *theunit, int flip, char  *item)
     theunit->factor *= number;
   else
     theunit->factor /= number;
-  item += len;
   while(*item){
     if (*item != '|') {
       fprintf(stderr,"Mysterious number error, should not happen\n");
       return 1;
     }
     item++;
-    sscanf(item,"%lf%n",&number,&len);
+    number = strtod(item, &item);
     if (!number) {
       zeroerror();
       return  1;
@@ -781,7 +780,6 @@ addnumber(struct unittype *theunit, int flip, char  *item)
       theunit->factor /= number;
     else
       theunit->factor *= number;
-    item+=len;
   }
   return 0;
 }
@@ -903,7 +901,7 @@ handlesum(struct unittype *theunit, char *toadd, int flip)
   char *first, *text, *rest;
   int retval;
   initializeunit(&firstunit);
-  first=strdup(toadd);
+  first=dupstr(toadd);
   rest=strchr(first,'+');   
   if (rest) *rest++=0;      
   if (isblankstr(first)){
@@ -1118,33 +1116,61 @@ int
 showanswer(char *havestr,struct unittype *have,
            char *wantstr,struct unittype *want)
 {
+   struct unittype invhave;
+   int doingrec;  /* reciprocal conversion? */
+   char *sep, *right, *left;
+
+   doingrec=0;
    havestr = removepadding(havestr);
    wantstr = removepadding(wantstr);
    if (compareunits(have, want)) {
-      printf("conformability error\n");
-      if (verbose) printf("\t%s = ",havestr);
-      else putchar('\t');
-      showunit(have);
-      if (verbose) printf("\t%s = ",wantstr);
-      else putchar('\t');
-      showunit(want);
-      return -1;
-   } else {
-      if (verbose) printf("\t%s = ",havestr);
-      else printf("\t* ");
-      printf(numformat, have->factor / want->factor);
-      if (verbose) {
-        if (numlength(wantstr)>0) printf(" *");
-        printf(" %s\n\t%s = (1 / ",wantstr,havestr);
-      } else printf("\n\t/ ");
-      printf(numformat, want->factor / have->factor);
-      if (verbose){
-        printf(")");
-        if (numlength(wantstr)>0) printf(" *");
-        printf(" %s",wantstr);
-      }
-      putchar('\n');
-   }
+        char **src,**dest;
+
+        invhave.factor=1/have->factor;
+        for(src=have->numerator,dest=invhave.denominator;*src;src++,dest++)
+           *dest=*src;
+        *dest=0;
+        for(src=have->denominator,dest=invhave.numerator;*src;src++,dest++)
+           *dest=*src;
+        *dest=0;
+        if (strictconvert || compareunits(&invhave, want)){
+	  printf("conformability error\n");
+	  if (verbose) printf("\t%s = ",havestr);
+	  else putchar('\t');
+	  showunit(have);
+	  if (verbose) printf("\t%s = ",wantstr);
+	  else putchar('\t');
+	  showunit(want);
+	  return -1;
+        }
+        printf("\treciprocal conversion\n");
+        have=&invhave;
+        doingrec=1;
+   } 
+   if (verbose) {
+     if (numlength(wantstr)>0) 
+       sep=" *";
+     else 
+       sep="";
+     if (!doingrec) 
+       left=right="";
+     else if (strchr(havestr,'/')) {
+       left="1 / (";
+       right=")";
+     } else {
+       left="1 / ";
+       right="";
+     }
+   }   
+   if (verbose) printf("\t%s%s%s = ",left,havestr,right);
+   else printf("\t* ");
+   printf(numformat, have->factor / want->factor);
+   if (verbose) 
+     printf("%s %s\n\t%s%s%s = (1 / ",sep,wantstr,left,havestr,right);
+   else printf("\n\t/ ");
+   printf(numformat, want->factor / have->factor);
+   if (verbose) printf(")%s %s",sep,wantstr);
+   putchar('\n');
    return 0;
 }
 
@@ -1206,6 +1232,7 @@ usage()
     -o, --output-format specify printf numeric output format\n\
     -q, --quiet         supress prompting\n\
         --silent        same as --quiet\n\
+    -s, --strict        suppress reciprocal unit conversion\n\
     -v, --verbose       print slightly more verbose output\n\
     -V, --version       print version number and exit\n", stdout);
    exit(3);
@@ -1309,7 +1336,7 @@ getuser(char **buffer, int *bufsize, char *query)
 #endif /* READLINE */
 
 
-char *shortoptions = "Vvqechf:o:";
+char *shortoptions = "Vvqechsf:o:";
 
 struct option longoptions[] = {
   {"version", no_argument, 0, 'V'},
@@ -1323,6 +1350,7 @@ struct option longoptions[] = {
   {"file", required_argument, 0, 'f'},
   {"output-format", required_argument, 0, 'o'},
   {"help",no_argument,0,'h'},
+  {"strict",no_argument,&strictconvert, 1},
   {0,0,0,0} };
 
 /* Process the args.  Returns 1 if interactive mode is desired, and 0
@@ -1354,7 +1382,10 @@ processargs(int argc, char **argv, char **from, char **to)
 	    break;
 	 case 'q':
 	    quiet = 1;
-	    break;
+	    break; 
+         case 's':
+            strictconvert = 1;
+            break;
          case 'v':
             verbose = 1;
             break;
